"""
@author: Payam Dibaeinia
"""

import os


def make_dir(path, writable = False):
    """
    writable: determines if the given path points to a readable/writable file, if
    so, only the intermidiate folders are made
    """
    if os.path.exists(path):
        print("directory " + path + " already exists", flush = True)

    if writable:
        path_list = path.split('/')[:-1]

        if path[0] == '/':
            path = '/' + '/'.join(path_list)
        else:
            path = '/'.join(path_list)

    if not os.path.exists(path):
        os.makedirs(path)



"""
@author: Payam Dibaeinia
"""

import numpy as np
import pandas as pd
from collections import defaultdict
import os


class gtSampler (object):

    def __init__(self, nTree, nSample, k, replacement = True, missingID = None):
        """
        nTree and nSample are assumed to be a python list
        """
        self.nTree = nTree
        self.nSample = nSample
        self.replacement = replacement
        self.k = k
        self.misID = missingID
        if isinstance(missingID, np.ndarray):
            self.compID = np.setdiff1d(range(k),missingID, True)

    def create_samples(self, path_read, path_write):
        """
        path_read: path to the input gene tree to create samples from
        path_write: path to the ouput directory were sampled trees will be written
        """
        last_ID = 0
        for curr_nSample, curr_nTree in zip (self.nSample, self.nTree):
            self._sample(curr_nSample, curr_nTree, last_ID, path_read, path_write)
            last_ID += curr_nSample



    def _sample (self, num_samples, tree_per_sample, last_sample_ID, path_read, path_write):
        """
        path_read: path to the input gene tree to create samples from
        path_write: path to the ouput directory were sampled trees will be written
        """

        for s in range(num_samples):
            os.mkdir(path_write + '/Sample_' + str(s + last_sample_ID))
            gtIDs = self._getTrIDs(tree_per_sample)
            with open(path_write + '/Sample_' + str(s + last_sample_ID) + '/sampledGeneTrees', 'w') as fw:
                with open(path_read) as fr:
                    for lID, l in enumerate(fr):
                        if lID in gtIDs:
                            fw.write(l.strip())
                            fw.write('\n')

    def _getTrIDs(self,size):
        if not isinstance(self.misID, np.ndarray):
            return np.random.choice(self.k, size, replace = self.replacement)

        else:
            nComp = int(len(self.compID) * size / self.k)
            nInComp = size - nComp
            compIDs = np.random.choice(self.compID, nComp, replace = self.replacement)
            inCompIDs = np.random.choice(self.misID, nInComp, replace = self.replacement)
            ret = np.concatenate((compIDs, inCompIDs))
            return ret.flatten()




"""
@author: Payam Dibaeinia
"""

import numpy as np
import shutil
from FASTRAL.Sampler import gtSampler
import time
import pandas as pd
import os


class FASTRAL (object):

    def __init__(self, flags):

        nt = [int(i) for i in flags.nt.split(',')]
        ns = [int(i) for i in flags.ns.split(',')]

        self.flags_ = flags
        if flags.incomp_id != None:
            incomp_id = pd.read_csv(flags.incomp_id, header = None, index_col = None)
            incomp_id = incomp_id.values.flatten()
        else:
            incomp_id = None

        print("START BUILDING SAMPLES ... ", flush=True)
        sampler = gtSampler(nTree = nt, nSample = ns, k = flags.k, replacement = flags.rep, missingID = incomp_id)
        sampler.create_samples(path_read = flags.it, path_write = flags.os)

        self.path_samples = self.flags_.os + '/Sample_'
        self.nTotalS_ = np.sum(ns)

        self.multi = None
        if self.flags_.multi:
            self.multi = ' -a {}'.format(self.flags_.multi)

    def run(self):
        t1 = time.time()
        self._run_ASTRID()
        t2 = time.time()

        self._aggregate_ASTRID_trees()

        t3 = time.time()
        self._run_ASTRAL()
        t4 = time.time()

        """
        write running times
        """
        header = ['ASTRID_time', 'ASTRAL_time', 'total_time']
        df = pd.DataFrame([[t2-t1, t4-t3, t2-t1 + t4-t3]])
        df.to_csv(self.flags_.time, header = header, sep = '\t', index = False)


    def _run_ASTRID(self):

        print("START RUNNING ASTRID ... ", flush=True)
        cline = self.flags_.path_ASTRID + ' -i ' + self.path_samples




        for s in range(self.nTotalS_):
            curr_cline = cline + str(s) + '/sampledGeneTrees'
            if self.multi:
                curr_cline += self.multi + ' -o ' + self.path_samples + str(s) + '/ASTRID_species_tree_' + str(s)
            else:
                curr_cline += ' -o ' + self.path_samples + str(s) + '/ASTRID_species_tree_' + str(s)

            print("     Running ASTRID on sample " + str(s), flush=True)
            status = os.system(curr_cline)
            if status < 0:
                raise ValueError('ASTRID was not run successfully')

    def _aggregate_ASTRID_trees(self):

        print("START AGGREGATING ASTRID's OUTPUTS ... ", flush=True)

        with open(self.flags_.aggregate,'wb') as wf:
            for s in range(self.nTotalS_):
                path = self.path_samples + str(s) +'/ASTRID_species_tree_' + str(s)
                with open(path,'rb') as rf:
                    shutil.copyfileobj(rf, wf)

    def _run_ASTRAL(self):

        cline = 'java -jar ' + self.flags_.path_ASTRAL + ' -i ' + self.flags_.it + ' -f ' + self.flags_.aggregate + ' -p ' + str(self.flags_.heuristics) + ' -o ' + self.flags_.o
        if self.multi:
            cline += self.multi
        print("START RUNNING ASTRAL ... ", flush=True)
        status = os.system(cline)
        if status < 0:
            raise ValueError('ASTRID was not run successfully')



"""
@author: Payam Dibaeinia
"""

import argparse
from FASTRAL.fastral import FASTRAL
from FASTRAL.utils import make_dir
import os
import warnings
import subprocess
import site
from pathlib import Path
import stat
import platform






def main():
    #find OS
    if platform.system() == "Darwin":
        sys = "osx"
    elif platform.system() == "Linux":
        sys = "linux"
    else:
        raise ValueError("Operating system is not supported. FASTRAL only runs on Linux and mac OS.")

    sitePackages = site.getsitepackages()[0]
    parser = argparse.ArgumentParser()

    parser.add_argument('--ns', type=str, default=None, help='number of samples, input multiple sample sizes with comma separated values', required = True)
    parser.add_argument('--nt', type=str, default=None, help='number of trees per sample, if there are multiple samples use comma separated values', required = True)
    parser.add_argument('--k', type=int, default=None, help='total number of input gene trees', required = True)
    parser.add_argument('--it', type=str, default=None, help='path to input gene trees', required = True)
    parser.add_argument('--os', type=str, default=None, help='path to samples folders to write sampled trees', required = True)
    parser.add_argument('--aggregate', type=str, default=None, help='path to write aggregated species trees', required = True)
    parser.add_argument('--o', type=str, default=None, help='path to write FASTRAL output species tree', required = True)
    parser.add_argument('--time', type=str, default=None, help='path to write running times', required = True)
    parser.add_argument('--rep', action='store_true', help='whether draw samples with replacement | Default: False', required = False)
    parser.add_argument('--path_ASTRID', type=str, default=sitePackages + '/FASTRAL/ASTRID/ASTRID-'+sys, help='path to ASTRID | Default: ASTRID-2 (version untagged-fdc5326080d364b87c5a) is used (see: https://github.com/pranjalv123/ASTRID/releases/tag/untagged-fdc5326080d364b87c5a)', required = False)
    parser.add_argument('--path_ASTRAL', type=str, default=sitePackages + '/FASTRAL/ASTRAL-modified/astral.5.7.3.modified.jar', help='path to ASTRAL | Default: modified ASTRAL 5.7.3 is used', required = False)
    parser.add_argument('--heuristics', type=int, default=0, help='heuristics level of ASTRAL | Default: 0', required = False)
    parser.add_argument('--multi', type=str, default=None, help='if input gene trees contain multiple individuals, specify the path to the mapping file', required = False)
    parser.add_argument('--incomp_id', type=str, default=None, help='(Optional) Path to a file containing the IDs of incomplete gene trees. If specified, sampling step makes sure that each sub-sample contains complete gene trees.', required = False)

    FLAGS = parser.parse_args()

    if not os.path.exists(FLAGS.path_ASTRAL):
        warnings.warn("This is the first time you are using FASTRAL! Let's build dependencies...")
        # make executables:
        f = Path(FLAGS.path_ASTRID)
        f.chmod(f.stat().st_mode | stat.S_IEXEC)

        # build dependencies
        cm = "cd {}/FASTRAL/ASTRAL-modified/ \n sh make.sh".format(sitePackages)
        p = subprocess.Popen(cm, shell=True)
        (output, err) = p.communicate()
        p_status = p.wait()


    make_dir(FLAGS.os)
    make_dir(FLAGS.time, writable = True)
    make_dir(FLAGS.aggregate, writable = True)
    make_dir(FLAGS.o, writable = True)

    method = FASTRAL(FLAGS)
    method.run()

if __name__ == "__main__":
    main()



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

VERSION = '1.0.0'
PACKAGE_NAME = 'FASTRAL'
AUTHOR = 'Payam Dibaeinia'
AUTHOR_EMAIL = 'dibaein2@illinois.edu'
URL = 'https://github.com/PayamDiba/FASTRAL'

LICENSE = 'MIT License'
DESCRIPTION = 'A hybrid summary method for reconstructing species tree from gene trees.'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

INSTALL_REQUIRES = [
      'numpy',
      'pandas',
      'absl-py'
]

setup(name=PACKAGE_NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      include_package_data=True,
      packages=find_packages(),
      entry_points ={'console_scripts': [ 'fastral = FASTRAL.fastral_infer:main']},
      python_requires='>3.5.2',
      classifiers =(
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License", ),

      )
